"use strict";
/*--------------------------------------------------------------------------

@sinclair/typebox-codegen

The MIT License (MIT)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

---------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.ModelToTypeScript = void 0;
const formatter_1 = require("../common/formatter");
const compiler_1 = require("@sinclair/typebox/compiler");
const Types = require("@sinclair/typebox");
var ModelToTypeScript;
(function (ModelToTypeScript) {
    function Any(schema) {
        return 'any';
    }
    function Array(schema) {
        const items = Visit(schema.items);
        return `Array<${items}>`;
    }
    function Boolean(schema) {
        return 'boolean';
    }
    function BigInt(schema) {
        return 'bigint';
    }
    function Constructor(schema) {
        const params = schema.parameters.map((param) => Visit(param)).join(', ');
        const returns = Visit(schema.returns);
        return `(new (${params}) => ${returns})`;
    }
    function Date(schema) {
        return 'Date';
    }
    function Function(schema) {
        const params = schema.parameters.map((param) => Visit(param)).join(', ');
        const returns = Visit(schema.returns);
        return `((${params}) => ${returns})`;
    }
    function Integer(schema) {
        return 'number';
    }
    function Intersect(schema) {
        return schema.allOf.map((schema) => Visit(schema)).join(' & ');
    }
    function Literal(schema) {
        if (typeof schema.const === 'string') {
            return `'${schema.const}'`;
        }
        else {
            return `${schema.const}`;
        }
    }
    function Never(schema) {
        return 'never';
    }
    function Null(schema) {
        return 'null';
    }
    function String(schema) {
        return 'string';
    }
    function Number(schema) {
        return 'number';
    }
    function Object(schema) {
        const properties = globalThis.Object.entries(schema.properties)
            .map(([key, value]) => {
            return `${key}: ${Visit(value)}`;
        })
            .join(',\n');
        return `{\n${properties}\n}`;
    }
    function Promise(schema) {
        const item = Visit(schema.item);
        return `Promise<${item}>`;
    }
    function Record(schema) {
        for (const [key, value] of globalThis.Object.entries(schema.patternProperties)) {
            const type = Visit(value);
            if (key === '^(0|[1-9][0-9]*)$') {
                return `Record<number, ${type}>`;
            }
            else {
                return `Record<string, ${type}>`;
            }
        }
        throw Error('TypeBoxToTypeScript: Unreachable');
    }
    function Ref(schema) {
        return schema.$ref;
    }
    function This(schema) {
        return schema.$ref;
    }
    function Tuple(schema) {
        if (schema.items === undefined)
            return `[]`;
        const items = schema.items.map((schema) => Visit(schema)).join(', ');
        return `[${items}]`;
    }
    function UInt8Array(schema) {
        return `Uint8Array`;
    }
    function Undefined(schema) {
        return `undefined`;
    }
    function Union(schema) {
        return schema.anyOf.map((schema) => Visit(schema)).join(' | ');
    }
    function Unknown(schema) {
        return `unknown`;
    }
    function Void(schema) {
        return `void`;
    }
    function Visit(schema) {
        if (reference_map.has(schema.$id))
            return schema.$id;
        if (schema.$id !== undefined)
            reference_map.set(schema.$id, schema);
        if (Types.TypeGuard.TAny(schema))
            return Any(schema);
        if (Types.TypeGuard.TArray(schema))
            return Array(schema);
        if (Types.TypeGuard.TBoolean(schema))
            return Boolean(schema);
        if (Types.TypeGuard.TBigInt(schema))
            return BigInt(schema);
        if (Types.TypeGuard.TConstructor(schema))
            return Constructor(schema);
        if (Types.TypeGuard.TDate(schema))
            return Date(schema);
        if (Types.TypeGuard.TFunction(schema))
            return Function(schema);
        if (Types.TypeGuard.TInteger(schema))
            return Integer(schema);
        if (Types.TypeGuard.TIntersect(schema))
            return Intersect(schema);
        if (Types.TypeGuard.TLiteral(schema))
            return Literal(schema);
        if (Types.TypeGuard.TNever(schema))
            return Never(schema);
        if (Types.TypeGuard.TNull(schema))
            return Null(schema);
        if (Types.TypeGuard.TNumber(schema))
            return Number(schema);
        if (Types.TypeGuard.TObject(schema))
            return Object(schema);
        if (Types.TypeGuard.TPromise(schema))
            return Promise(schema);
        if (Types.TypeGuard.TRecord(schema))
            return Record(schema);
        if (Types.TypeGuard.TRef(schema))
            return Ref(schema);
        if (Types.TypeGuard.TThis(schema))
            return This(schema);
        if (Types.TypeGuard.TString(schema))
            return String(schema);
        if (Types.TypeGuard.TTuple(schema))
            return Tuple(schema);
        if (Types.TypeGuard.TUint8Array(schema))
            return UInt8Array(schema);
        if (Types.TypeGuard.TUndefined(schema))
            return Undefined(schema);
        if (Types.TypeGuard.TUnion(schema))
            return Union(schema);
        if (Types.TypeGuard.TUnknown(schema))
            return Unknown(schema);
        if (Types.TypeGuard.TVoid(schema))
            return Void(schema);
        return 'unknown';
    }
    function GenerateType(model, $id) {
        reference_map.clear();
        const type = model.types.find((type) => type.$id === $id);
        if (type === undefined)
            return `export type ${$id} = unknown`;
        return `export type ${type.$id} = ${Visit(type)}`;
    }
    ModelToTypeScript.GenerateType = GenerateType;
    const reference_map = new Map();
    function Generate(model) {
        reference_map.clear();
        const definitions = [];
        for (const type of model.types) {
            const definition = `export type ${type.$id} = ${Visit(type)}`;
            const assertion = `export const ${type.$id} = (() => { ${compiler_1.TypeCompiler.Code(type, model.types, { language: 'typescript' })} })();`;
            const rewritten = assertion.replaceAll(`return function check(value: any): boolean`, `return function check(value: any): value is ${type.$id}`);
            definitions.push(`
      ${definition}
      ${rewritten}
      `);
        }
        const output = [...definitions];
        return formatter_1.Formatter.Format(output.join('\n\n'));
    }
    ModelToTypeScript.Generate = Generate;
})(ModelToTypeScript || (exports.ModelToTypeScript = ModelToTypeScript = {}));
